#include "Preprocessing.h"

CPreProcessing::CPreProcessing(int label_num, int max)
{
   int i,j,k,l;
   struct Conflict *node;

   labelNum = label_num;
   max_k=max;
   quartetNum= (labelNum)*(labelNum-1)/2*(labelNum-2)/3*(labelNum-3)/4;

   quartetTable = new CQuartetTable(labelNum);
 
   errorUpperBound = (int **)malloc( (labelNum+1)*sizeof(int *) );
   for (i=1;i<=labelNum;i++)
      errorUpperBound[i]=(int *)calloc( (labelNum+1), sizeof(int) );
      
    weightTable = (int ****)malloc( (labelNum+1)*sizeof(int ***) );
    for (i=1;i<=labelNum;i++)
      weightTable[i]=(int ***)malloc( (labelNum+1)*sizeof(int **) );
    for (i=1;i<=labelNum;i++)
      for (j=i+1;j<=labelNum;j++)
         weightTable[i][j]=(int **)malloc( (labelNum+1)*sizeof(int *) );
    for (i=1;i<=labelNum;i++)
      for (j=i+1;j<=labelNum;j++)
        for (k=i+1;k<=labelNum;k++)
	       weightTable[i][j][k]=(int *)calloc( labelNum+1, sizeof(int) );
		   
    head=(struct Conflict *)calloc(1, sizeof( struct Conflict ) );
	head->q1[0]=0;
	head->next=0;
}

CPreProcessing::~CPreProcessing()
{
}

void CPreProcessing::SetStartTime(time_t t)
{
	start=t;
}


void CPreProcessing::PreProcessing(char *filename)
{
    int i,j;
	
    GenerateConflictList();

  
    //// errorUpperBound: the maximum number of  errors including (i,j)
    //// at most we need change max_k quartets,
    //// the quartet errors not include i,j is at least LowerBoundNotInclude(i,j)
    ///  Therefore, we need change at most max_k-LowerBoundNotInclude(i,j) number
    //   of quartets including (i,j).
    for (i=1;i<=labelNum-1;i++)
	  for (j=i+1;j<=labelNum;j++)
	  	errorUpperBound[i][j]=max_k-LowerBoundNotInclude(i,j);  
   // printf("errorUpperBound:\n");   
   // DisplayMatrix(errorUpperBound); 
    
	

    DomainReduction();
    
    GenerateSmodelSource(filename);
//	printf("preprocessing finished");
}

int CPreProcessing::LowerBoundNotInclude(UNSIGNCHAR a, UNSIGNCHAR  b)
{
    int *conflict_count;
    int lowerbound,count=0,loc;
    int i,j,k,l;
    int conflicts_not_include=0;
    struct Conflict *node=head->next;

    for (i=1;i<=labelNum-1;i++)
     for (j=i+1;j<=labelNum;j++)
       for (k=i+1;k<=labelNum-1;k++)
         for (l=k+1;l<=labelNum;l++)
            weightTable[i][j][k][l]=0;
    
    conflict_count =(int *)calloc( quartetNum , sizeof(int));
        
    while (node!=0)
    {
       if ( (node->q1[0]==a) || (node->q1[1]==a) || (node->q1[2]==a) ||(node->q1[3]==a) ||
            (node->q2[0]==a) || (node->q2[1]==a) || (node->q2[2]==a) ||(node->q2[3]==a) ||
            (node->q3[0]==a) || (node->q3[1]==a) || (node->q3[2]==a) ||(node->q3[3]==a) ||
            (node->q1[0]==b) || (node->q1[1]==b) || (node->q1[2]==b) ||(node->q1[3]==b) ||
            (node->q2[0]==b) || (node->q2[1]==b) || (node->q2[2]==b) ||(node->q2[3]==b) ||
            (node->q3[0]==b) || (node->q3[1]==b) || (node->q3[2]==b) ||(node->q3[3]==b) )
       {
            node=node->next;
            continue;
       } 
       conflicts_not_include++; 
       
       loc = weightTable[node->q1[0]][node->q1[1]][node->q1[2]][node->q1[3]];
       if (loc>0)
         conflict_count[loc]++;
       else
       {
         conflict_count[count]=1;
         weightTable[node->q1[0]][node->q1[1]][node->q1[2]][node->q1[3]]=count;
         count++;
       }
       loc = weightTable[node->q2[0]][node->q2[1]][node->q2[2]][node->q2[3]];
       if (loc>0)
         conflict_count[loc]++;
       else
       {
         conflict_count[count]=1;
          weightTable[node->q2[0]][node->q2[1]][node->q2[2]][node->q2[3]]=count;
         count++;
       }
       loc = weightTable[node->q3[0]][node->q3[1]][node->q3[2]][node->q3[3]];
       if (loc>0)
         conflict_count[loc]++;
       else
       {
         conflict_count[count]=1;
         weightTable[node->q3[0]][node->q3[1]][node->q3[2]][node->q3[3]]=count;
         count++;
       }
       node=node->next;
    } 
    

    QuickSort(conflict_count,0,count);

    lowerbound=0;
    while(lowerbound< conflicts_not_include)
    {
        conflicts_not_include-=conflict_count[lowerbound];
        lowerbound++;
    }
    return lowerbound; 
              
}

void swap(int *a, int *b)
{
     int tmp;
	 tmp=*a;
	 *a=*b;
	 *b=tmp;
}


void CPreProcessing::QuickSort(int *cost,  int low, int high)
{
	UNSIGNCHAR pivot;

	int ScanUp, ScanDown;
	int mid;

	if (high-low<=0 ) 
		return;
	else
		if (high-low==1)
		{
			if (cost[high] > cost[low])
			{
				swap(cost+low, cost+high);
				
			}
			return;
		}
     
	mid=(low+high)/2;
	pivot=cost[mid];


	swap(cost+mid,cost+low);

	ScanUp=low+1;
	ScanDown=high;

	do
	{
		while (ScanUp<=ScanDown && cost[ScanUp]>=pivot)
			ScanUp++;
		while (pivot > cost[ScanDown])
			ScanDown--;
		if (ScanUp<ScanDown)
		{
			swap(cost+ScanUp, cost+ScanDown);

		}
	}
	while (ScanUp <ScanDown);

	cost[low]=cost[ScanDown];
	cost[ScanDown]=pivot;

	if (low < ScanDown-1)
		QuickSort(cost,  low, ScanDown-1);
	if (ScanDown+1 < high)
        QuickSort(cost,  ScanDown+1, high);
}

void  CPreProcessing::AddConflict(UNSIGNCHAR *q1,UNSIGNCHAR *q2,UNSIGNCHAR *q3)
{
    int i;
    struct Conflict *node;
    node=(struct Conflict *)calloc(1, sizeof( struct Conflict ) );
    
    for (i=0;i<4;i++)
    {
        node->q1[i]=q1[i];
        node->q2[i]=q2[i];
        node->q3[i]=q3[i];
    }
    node->next=head->next;
	head->next=node;
}      


///////in the preprocessing, generate the local conflict list
void CPreProcessing::GenerateConflictList()
{
   int  conflict_count=0, total=0;
	UNSIGNCHAR i,j,k,l,m;
	UNSIGNCHAR q1[4],q2[4],q3[4],q4[4],q5[4];

     
    for (i=1;i<=labelNum-4;i++)
	  for (j=i+1;j<=labelNum-3;j++)
	    for (k=j+1;k<=labelNum-2;k++)
	      for (l=k+1;l<=labelNum-1;l++)
	        for (m=l+1;m<=labelNum;m++)
	        {
                if (quartetTable->find(i, j, k, l,  q1)==0)
					printf("find error!!(%d,%d,%d,%d)\n",i,j,k,l);
           	    if (quartetTable->find(i, j, k, m,  q2)==0)
					printf("find error!!(%d,%d,%d,%d)\n",i,j,k,m);
			    if (quartetTable->find(i, j, l, m,  q3)==0)
					printf("find error!!(%d,%d,%d,%d)\n",i,j,l,m);
			    if (quartetTable->find(i, k, l, m,  q4)==0)
					printf("find error!!(%d,%d,%d,%d)\n",i,k,l,m);
			    if (quartetTable->find(j, k, l, m, q5)==0)
					printf("find error!!(%d,%d,%d,%d)\n",j,k,l,m);

		    	if (Compatible(q1,q2,q3)==0)
		    	{
					AddConflict(q1,q2,q3);
			    	conflict_count++;
				}
				if (Compatible(q1,q2,q4)==0)
		    	{
                    AddConflict(q1,q2,q4);
                    conflict_count++;
				}
				if (Compatible(q1,q2,q5)==0)
		    	{
                    AddConflict(q1,q2,q5);					
                    conflict_count++;
				}
				if (Compatible(q1,q3,q4)==0)
		    	{
                    AddConflict(q1,q3,q4);
                    conflict_count++;
				}
				if (Compatible(q1,q3,q5)==0)
		    	{
                    AddConflict(q1,q3,q5);
                    conflict_count++;
				}
				if (Compatible(q1,q4,q5)==0)
		    	{
                    AddConflict(q1,q4,q5);
					conflict_count++;
				}
	            if (Compatible(q2,q3,q4)==0)
		    	{
                    AddConflict(q2,q3,q4);
                    conflict_count++;
				}
	            if (Compatible(q2,q3,q5)==0)
		    	{
                    AddConflict(q2,q3,q5);
					conflict_count++;
				}
	            if (Compatible(q2,q4,q5)==0)
		    	{
                    AddConflict(q2,q4,q5);
                    conflict_count++;
				}
	            if (Compatible(q3,q4,q5)==0)
		    	{
                    AddConflict(q3,q4,q5);
					conflict_count++;
				}
			    total++;
			 }
    printf("total %d local subsets, conflit %d\n", total*10, conflict_count);
    totalConflicts=conflict_count;
  
}


int CPreProcessing::ExchangeableCheck(UNSIGNCHAR i,UNSIGNCHAR j,UNSIGNCHAR quartet1[4], UNSIGNCHAR quartet2[4])
{
	UNSIGNCHAR q[4];
	UNSIGNCHAR k,a,b,c,d;

	///// i is in quartet1, j is in quartet 2.
	/////step 1: change the i in quartet1 to j, and store the new quartet in q
	for (k=0;k<4;k++)
	   if (quartet1[k]==i)
		   q[k]=j;
		else
		   q[k]=quartet1[k];
    //// step 2: change the new quaret q into the standard form
    a=q[0];b=q[1];c=q[2];d=q[3];
    quartetTable->formquartet(a,b,c,d,q);

    ///step 3: check whether q is same as quartet2
    for (k=0;k<4;k++)
      if (q[k]!=quartet2[k])
        return 0;

    return 1;

}

void CPreProcessing:: DomainReduction()
{
   int **exchangeable, **nonexchangeable;
   int **consistent, **inconsistent;

   
   
   int exp2[9];
   int k_bar, p1_lower , p1_upper ,p2_lower, p2_upper,p3_lower , p4_upper;
   int diff1, diff2, diff3, diff4, diff_conf, diff_exg;
   
   CBitLabelSet **between;

   UNSIGNCHAR quartet1[4],quartet2[4],q[4];
   UNSIGNCHAR a,b,c,d;
   UNSIGNCHAR i,j,k,l,m,n;

   if (labelNum %2==0)
      max_height=labelNum/2;
   else
      max_height=labelNum/2+1;

///// alloc memories
    exchangeable = (int **)malloc( (labelNum+1)*sizeof(int *) );
    for (i=1;i<=labelNum;i++)
      exchangeable[i]=(int *)calloc( (labelNum+1), sizeof(int) );

    nonexchangeable = (int **)malloc( (labelNum+1)*sizeof(int *) );
    for (i=1;i<=labelNum;i++)
      nonexchangeable[i]=(int *)calloc( (labelNum+1), sizeof(int) );

    consistent = (int **)malloc( (labelNum+1)*sizeof(int *) );
    for (i=1;i<=labelNum;i++)
      consistent[i]=(int *)calloc( (labelNum+1), sizeof(int) );

    inconsistent = (int **)malloc( (labelNum+1)*sizeof(int *) );
    for (i=1;i<=labelNum;i++)
      inconsistent[i]=(int *)calloc( (labelNum+1), sizeof(int) );

    domain = (int **)malloc( (labelNum+1)*sizeof(int *) );
    for (i=1;i<=labelNum;i++)
      domain[i]=(int *)calloc( (labelNum+1), sizeof(int) );

    ultra = (int **)malloc( (labelNum+1)*sizeof(int *) );
    for (i=1;i<=labelNum;i++)
      ultra[i]=(int *)calloc( (labelNum+1), sizeof(int) );

    childrenlist = (int **)malloc( (labelNum+1)*sizeof(int *) );
    for (i=1;i<=labelNum;i++)
      childrenlist[i]=(int *)calloc( (labelNum+1), sizeof(int) );

    between = (CBitLabelSet **)malloc( (labelNum+1)*sizeof(CBitLabelSet *) );
    for (i=1;i<=labelNum;i++)
    {
       between[i]=(CBitLabelSet *)calloc( (labelNum+1), sizeof(CBitLabelSet) );
       for (j=1;j<=labelNum;j++)
         between[i][j].AllocMem(labelNum);
    }

    
   /// calculate the conflicting and nonconflicting numbers for 
   /// each pair i,j
    for (i=1;i<=labelNum-3;i++)
      for (j=i+1;j<=labelNum-2;j++)
        for (k=j+1;k<=labelNum-1;k++)
          for (l=k+1;l<=labelNum;l++)
	      {
    		  if (quartetTable->find(i, j, k, l,  quartet1)==0)
                  printf("1error!!(%d,%d,%d,%d)\n",i,j,k,l);

                  a=quartet1[0];
                  b=quartet1[1];
                  c=quartet1[2];
                  d=quartet1[3];

                  consistent[a][b]++;
		  consistent[b][a]++;
		  consistent[c][d]++;
		  consistent[d][c]++;

		  inconsistent[a][c]++;
		  inconsistent[c][a]++;
		  inconsistent[a][d]++;
		  inconsistent[d][a]++;

		  inconsistent[b][c]++;
		  inconsistent[c][b]++;
		  inconsistent[b][d]++;
		  inconsistent[d][b]++;
              }


    ////// calculate the exchangeable and nonexchangeable numbers for 
    /////  each pair i,j
	for (i=1;i<=labelNum-1;i++)
	  for (j=i+1;j<=labelNum;j++)
          {
     	    for (k=1;k<=labelNum-2;k++)
            {
              if ((k==i)||(k==j)) continue;
              for (l=k+1;l<=labelNum-1;l++)
              {
                if ((l==i)||(l==j)) continue;
                for (m=l+1;m<=labelNum;m++)
                {
                   if ((m==i)||(m==j)) continue;
		   quartetTable->formquartet(i, k, l, m,  q);

		   if (quartetTable->find(q,  quartet1)==0)
		     printf("2error!!(%d,%d,%d,%d)\n",i,k,l,m);

		    quartetTable->formquartet(j, k, l, m,  q);
		    if (quartetTable->find( q,  quartet2)==0)
		      printf("3error!!(%d,%d,%d,%d)\n",j,k,l,m);
		    if (ExchangeableCheck(i,j,quartet1, quartet2)==1)
                    {
		        exchangeable[i][j]++;
		        exchangeable[j][i]++;
                    }
		    else
                    {
                       between[i][j].Add(k); //// the labels between i, j
                       between[i][j].Add(l);
                       between[i][j].Add(m);
		       nonexchangeable[i][j]++;
		       nonexchangeable[j][i]++;
                    }
                 }///for m
              } /// for l
            }///for k
	  }//for j

//   printf("exchangeable:\n");
//   DisplayMatrix(exchangeable);

//   printf("nonexchangeable:\n");
//   DisplayMatrix(nonexchangeable);

//   printf("consistent:\n");
//   DisplayMatrix(consistent);

///   printf("inconsistent:\n");
///   DisplayMatrix(inconsistent);

   n=labelNum;
/////sibling check
    for  (i=1;i<=labelNum-1;i++)
      for (j=i+1;j<=labelNum;j++)
      {
            if (2*inconsistent[i][j]+nonexchangeable[i][j]<= n-3)
                //between[i][j].GetSize()*(between[i][j].GetSize()-1)/2)
            {
                domain[i][j]=1;
                domain[j][i]=1;
   //             printf("(%d,%d) siblings\n",i,j);
            }
      }


  /// the domain of i, j (i<j) is [ domain[j][i], domain[i][j] ]   

   int bound;
   int min_height=1;
   
   max_height=n/2;


   //// exp2[k]=2^k 
   exp2[1]=2;  exp2[2]=4;   exp2[3]=8;   exp2[4]=16;
   exp2[5]=32; exp2[6]=64;  exp2[7]=128; exp2[8]=256;
   
   for  (i=1;i<=labelNum-1;i++)
     for (j=i+1;j<=labelNum;j++)
     {
          bound = errorUpperBound[i][j];
            
          for (k=1 ; k<=max_height; k++)
	      {
              //////k_bar= min{n-k-2, 2^k}  
              k_bar= n-k-2;
              if (k_bar > exp2[k]-2) k_bar=exp2[k]-2;
              
              ////p1_lower = {n-2-k_bar \choose 2}               //// lower bound of consistent
              /// p1_upper = {k-1 \choose 2} +{n-k-1 \choose 2}  /// upper bound of consistent
              /// p2_lower = (k-1)(n-k-1)                        ///lower bound of inconsistent
              /// p2_upper = {k_bar \choose 2}+ k_bar (n-2-k_bar) ///upper bound of inconsistent
              /// p3_lower = {n-2-k_bar \choose 3}                /////lower bound of exchangeable
              /// p4_upper = {k_bar \choose 3} +{k_bar \choose 2} (n-2-k_bar) /// upper bound of nonexchangeable
              
              p1_lower = (n-2-k_bar)*(n-3-k_bar)/2;
              if (p1_lower<0) p1_lower=0;
              
              if (k<=2) 
                 p1_upper =(n-k-1)*(n-k-2)/2;
              else
                 p1_upper = (k-1)*(k-2)/2+(n-k-1)*(n-k-2)/2;
              
              p2_lower =(k-1)*(n-k-1);
              
              if (k==1)
                p2_upper=0;
              else
                p2_upper = k_bar*(k_bar-1)/2+k_bar*(n-2-k_bar);
              
              
              p3_lower = (n-2-k_bar)*(n-3-k_bar)*(n-4-k_bar)/6;
              
              p4_upper = k_bar*(k_bar-1)*(k_bar-2)/6+ (k_bar*(k_bar-1)/2)*(n-2-k_bar);
          
              ///// calculate the difference
              if (consistent[i][j]< p1_lower) 
                diff1=  p1_lower-consistent[i][j];
              else if (consistent[i][j]> p1_upper) 
                diff1= consistent[i][j]- p1_upper;
              else diff1=0;
                
              if (inconsistent[i][j]< p2_lower) 
                diff2=  p2_lower-inconsistent[i][j];
              else if (inconsistent[i][j]> p2_upper) 
                diff2= inconsistent[i][j]- p2_upper;
              else diff2=0;
            
              if (exchangeable[i][j]<p3_lower)
                 diff3 = p3_lower- exchangeable[i][j];
              else
                 diff3 =0;
              
              if (nonexchangeable[i][j]>p4_upper)
                 diff4 = nonexchangeable[i][j]-p4_upper ;
              else
                 diff4 =0;  
                 
     /*       if ((i==3)&& (j==4)&&(k==1))
            {
              printf("consist=%d, inconsist=%d, exchang=%d, nonexg=%d, \n",
              consistent[i][j], inconsistent[i][j], exchangeable[i][j], nonexchangeable[i][j]);  
              
              printf("p1_lower=%d, p1_upper=%d, p2_lower=%d,p2_upper=%d, p3_lower=%d, p4_upper=%d \n",
              p1_lower, p1_upper, p2_lower, p2_upper, p3_lower,  p4_upper); 
                              
              printf("bound=%d, k=%d, diff1=%d, diff2=%d, diff3=%d, diff4=%d, diff=%d\n",
              errorUpperBound[i][j],k,diff1,diff2,diff3,diff4,diff1+diff2+diff3+diff4);   
            }
       */     
            if (diff1>diff2)
              diff_conf=diff1;
            else
              diff_conf=diff2;
            if (diff3>diff4)
              diff_exg=  diff3;
            else 
              diff_exg=diff4;
            
              if (diff_conf+diff_exg<=bound) //////  k is ok in domain of (i,j)
              {  
                 if (domain[j][i]==0)/// lower bound of domain is not set
                 {                   //// first set domain
                   domain[j][i]=k;
                  
                 }
                 else if (domain[i][j]!=1) ///lower bound of domain is set and not sibling
                   domain[i][j]=k; //// we set upper bound   
              }  
              else  if ((domain[j][i]>0) && (domain[i][j]!=1))  ///k is not ok, mark it
              { 
               
                domain[i][j]=-k;
                if (k>min_height)
                  min_height=k;
                break;
              }           
          }
        }

                
	                 
   // printf("domain:\n");
   // DisplayMatrix(domain);    
    

    //printf("min_height:=%d\n", min_height);



     	        
    for  (i=1;i<=labelNum-1;i++)
	  for (j=i+1;j<=labelNum;j++)
	  {
            if( (domain[i][j]==0) && (domain[j][i]==0) )
            {
                domain[j][i]=max_height;
                   domain[i][j]=max_height;
            }
            if  (domain[i][j]<0) 
            {
                if (domain[j][i]==0) /// LCA must be root
                {   
                   domain[j][i]=max_height;
                   domain[i][j]=max_height;
                }
                else /// LCA might be root
                {
                    if (-domain[i][j]< min_height-1) 
                      domain[i][j]=-domain[i][j]-1;
                    //else
                      //domain[i][j]=max_height;
                }
                
            }
  
      }
    
    best->Best_Qk();
    
    struct Partition  head_p[20]; 
    
    for (i=0;i<20;i++)
    {
      head_p[i].left=new CLabelSet(labelNum+1);
      head_p[i].left->Clear();
      head_p[i].chosen=0;
    }
    
    best->GetAllCandidateEdges(head_p);


    k=0;    
	while ((head_p[k].left !=0)&&(k<20))
	{
     //  head_p[k].left->Display(); 
       // printf("  k=%d,size=%d\n",k,head_p[k].left->GetSize());
        
        if (head_p[k].chosen==1)
        {
            
            

          for (i=0;i<head_p[k].left->GetSize()-1;i++)
             for (j=i+1;j<head_p[k].left->GetSize();j++)
             {
                    if (head_p[k].left->GetSize()< domain[head_p[k].left->Pop(i)][head_p[k].left->Pop(j)])
                 domain[head_p[k].left->Pop(i)][head_p[k].left->Pop(j)]= head_p[k].left->GetSize();
                    
             }		
       }

        k++;   
   }
    /*
     for  (i=1;i<=labelNum-1;i++)
	  for (j=i+1;j<=labelNum;j++)
      {
            if (domain[i][j]<0)
            {
         //     printf("(%d,%d)=[%d-%d,%d]\n",i,j,domain[j][i], -domain[i][j]-1,max_height);
              ///domain[i][j]=max_height;
            }
            else
           //   printf("(%d,%d)=[%d,%d]\n",i,j,domain[j][i], domain[i][j]);
      }*/      
}

void CPreProcessing::DisplayMatrix(int **m)
{
	int i,j;
    for (i=1;i<=labelNum;i++)
    {
	  	for (j=1;j<=labelNum;j++)
	  	  printf("%3d ", m[i][j]);
	  	printf("\n");
	}
	printf("\n\n");
}

void CPreProcessing::LoadFromFile(char *filename)
{
   labelNum=quartetTable->LoadFromFile(filename);
   
           
    best=new CBest(labelNum);
    best->LoadFromFile(filename, 1);
    if (labelNum<=10)
        best->Setm(2);
    else if (labelNum<=20)
        best->Setm(4);
    else
        best->Setm(6);
   
   
}

int CPreProcessing::LabelExist(int label, int *list)
{
	int i=0;
	while ( (i<5) && (list[i]>0) )
	{
		if (list[i]==label)
		  return 0;
		else
		  i++;
	}

	if (i<5)
	{
		list[i]=label;
		return 1;
	}
	else
	{
		printf("full!!!, label=%d, list=%d,%d,%d,%d,%d\n",
		       label, list[0],list[1],list[2],list[3],list[4]);
		return 0;
	}
}



//// q1[0], q1[1] | q1[2], q1[3]
//// q2[0], q2[1] | q2[2], q2[3]
//// q3[0], q3[1] | q3[2], q3[3]

////  there are 8 possible combinations

int CPreProcessing::Compatible(UNSIGNCHAR q1[4],UNSIGNCHAR q2[4],UNSIGNCHAR q3[4])
{
	int left_list[5],  right_list[5];
	int diff_left=0,  diff_right=0;
	int sig_left=8,sig_right;
	int i;

	for (i=0;i<5;i++)
	{
		left_list[i]=0;
		right_list[i]=0;
	}

//// q1[0], q1[1] | q1[2], q1[3]
//// q2[0], q2[1] | q2[2], q2[3]
//// q3[0], q3[1] | q3[2], q3[3]

        diff_left=0;  diff_right=0;
	diff_left+=LabelExist(q1[0],left_list);
	diff_left+=LabelExist(q1[1],left_list);
	diff_left+=LabelExist(q2[0],left_list);
	diff_left+=LabelExist(q2[1],left_list);
	diff_left+=LabelExist(q3[0],left_list);
	diff_left+=LabelExist(q3[1],left_list);

	diff_right+=LabelExist(q1[2],right_list);
	diff_right+=LabelExist(q1[3],right_list);
	diff_right+=LabelExist(q2[2],right_list);
	diff_right+=LabelExist(q2[3],right_list);
	diff_right+=LabelExist(q3[2],right_list);
	diff_right+=LabelExist(q3[3],right_list);

	if (diff_left<sig_left)
	{
		sig_left=diff_left;
		sig_right=diff_right;
	}


//// q1[0], q1[1] | q1[2], q1[3]
//// q2[0], q2[1] | q2[2], q2[3]
//// q3[2], q3[3] | q3[0], q3[1]
    diff_left=0;  diff_right=0;
	for (i=0;i<5;i++)
	{
		left_list[i]=0;
		right_list[i]=0;
	}

	diff_left+=LabelExist(q1[0],left_list);
	diff_left+=LabelExist(q1[1],left_list);
	diff_left+=LabelExist(q2[0],left_list);
	diff_left+=LabelExist(q2[1],left_list);
	diff_left+=LabelExist(q3[2],left_list);
	diff_left+=LabelExist(q3[3],left_list);

	diff_right+=LabelExist(q1[2],right_list);
	diff_right+=LabelExist(q1[3],right_list);
	diff_right+=LabelExist(q2[2],right_list);
	diff_right+=LabelExist(q2[3],right_list);
	diff_right+=LabelExist(q3[0],right_list);
	diff_right+=LabelExist(q3[1],right_list);

	if (diff_left<sig_left)
	{
		sig_left=diff_left;
		sig_right=diff_right;
	}

//// q1[0], q1[1] | q1[2], q1[3]
//// q2[2], q2[3] | q2[0], q2[1]
//// q3[0], q3[1] | q3[2], q3[3]
    diff_left=0;  diff_right=0;
	for (i=0;i<5;i++)
	{
		left_list[i]=0;
		right_list[i]=0;
	}

	diff_left+=LabelExist(q1[0],left_list);
	diff_left+=LabelExist(q1[1],left_list);
	diff_left+=LabelExist(q2[2],left_list);
	diff_left+=LabelExist(q2[3],left_list);
	diff_left+=LabelExist(q3[0],left_list);
	diff_left+=LabelExist(q3[1],left_list);

	diff_right+=LabelExist(q1[2],right_list);
	diff_right+=LabelExist(q1[3],right_list);
	diff_right+=LabelExist(q2[0],right_list);
	diff_right+=LabelExist(q2[1],right_list);
	diff_right+=LabelExist(q3[2],right_list);
	diff_right+=LabelExist(q3[3],right_list);

	if (diff_left<sig_left)
	{
		sig_left=diff_left;
		sig_right=diff_right;
	}


//// q1[0], q1[1] | q1[2], q1[3]
//// q2[2], q2[3] | q2[0], q2[1]
//// q3[2], q3[3] | q3[0], q3[1]
    diff_left=0;  diff_right=0;
	for (i=0;i<5;i++)
	{
		left_list[i]=0;
		right_list[i]=0;
	}
	diff_left+=LabelExist(q1[0],left_list);
	diff_left+=LabelExist(q1[1],left_list);
	diff_left+=LabelExist(q2[2],left_list);
	diff_left+=LabelExist(q2[3],left_list);
	diff_left+=LabelExist(q3[2],left_list);
	diff_left+=LabelExist(q3[3],left_list);

	diff_right+=LabelExist(q1[2],right_list);
	diff_right+=LabelExist(q1[3],right_list);
	diff_right+=LabelExist(q2[0],right_list);
	diff_right+=LabelExist(q2[1],right_list);
	diff_right+=LabelExist(q3[0],right_list);
	diff_right+=LabelExist(q3[1],right_list);

	if (diff_left<sig_left)
	{
		sig_left=diff_left;
		sig_right=diff_right;
	}

//// q1[2], q1[3] | q1[0], q1[1]
//// q2[0], q2[1] | q2[2], q2[3]
//// q3[0], q3[1] | q3[2], q3[3]
    diff_left=0;  diff_right=0;
    	for (i=0;i<5;i++)
	{
		left_list[i]=0;
		right_list[i]=0;
	}
    diff_left+=LabelExist(q1[2],left_list);
	diff_left+=LabelExist(q1[3],left_list);
	diff_left+=LabelExist(q2[0],left_list);
	diff_left+=LabelExist(q2[1],left_list);
	diff_left+=LabelExist(q3[0],left_list);
	diff_left+=LabelExist(q3[1],left_list);

	diff_right+=LabelExist(q1[0],right_list);
	diff_right+=LabelExist(q1[1],right_list);
	diff_right+=LabelExist(q2[2],right_list);
	diff_right+=LabelExist(q2[3],right_list);
	diff_right+=LabelExist(q3[2],right_list);
	diff_right+=LabelExist(q3[3],right_list);

	if (diff_left<sig_left)
	{
		sig_left=diff_left;
		sig_right=diff_right;
	}

//// q1[2], q1[3] | q1[0], q1[1]
//// q2[0], q2[1] | q2[2], q2[3]
//// q3[2], q3[3] | q3[0], q3[1]
    diff_left=0;  diff_right=0;
	for (i=0;i<5;i++)
	{
		left_list[i]=0;
		right_list[i]=0;
	}
	diff_left+=LabelExist(q1[2],left_list);
	diff_left+=LabelExist(q1[3],left_list);
	diff_left+=LabelExist(q2[0],left_list);
	diff_left+=LabelExist(q2[1],left_list);
	diff_left+=LabelExist(q3[2],left_list);
	diff_left+=LabelExist(q3[3],left_list);

	diff_right+=LabelExist(q1[0],right_list);
	diff_right+=LabelExist(q1[1],right_list);
	diff_right+=LabelExist(q2[2],right_list);
	diff_right+=LabelExist(q2[3],right_list);
	diff_right+=LabelExist(q3[0],right_list);
	diff_right+=LabelExist(q3[1],right_list);

	if (diff_left<sig_left)
	{
		sig_left=diff_left;
		sig_right=diff_right;
	}

//// q1[2], q1[3] | q1[0], q1[1]
//// q2[2], q2[3] | q2[0], q2[1]
//// q3[0], q3[1] | q3[2], q3[3]
    diff_left=0;  diff_right=0;
	for (i=0;i<5;i++)
	{
		left_list[i]=0;
		right_list[i]=0;
	}
	diff_left+=LabelExist(q1[2],left_list);
	diff_left+=LabelExist(q1[3],left_list);
	diff_left+=LabelExist(q2[2],left_list);
	diff_left+=LabelExist(q2[3],left_list);
	diff_left+=LabelExist(q3[0],left_list);
	diff_left+=LabelExist(q3[1],left_list);

	diff_right+=LabelExist(q1[0],right_list);
	diff_right+=LabelExist(q1[1],right_list);
	diff_right+=LabelExist(q2[0],right_list);
	diff_right+=LabelExist(q2[1],right_list);
	diff_right+=LabelExist(q3[2],right_list);
	diff_right+=LabelExist(q3[3],right_list);

	if (diff_left<sig_left)
	{
		sig_left=diff_left;
		sig_right=diff_right;
	}

//// q1[2], q1[3] | q1[0], q1[1]
//// q2[2], q2[3] | q2[0], q2[1]
//// q3[2], q3[3] | q3[0], q3[1]
    diff_left=0;  diff_right=0;
	for (i=0;i<5;i++)
	{
		left_list[i]=0;
		right_list[i]=0;
	}
	diff_left+=LabelExist(q1[2],left_list);
	diff_left+=LabelExist(q1[3],left_list);
	diff_left+=LabelExist(q2[2],left_list);
	diff_left+=LabelExist(q2[3],left_list);
	diff_left+=LabelExist(q3[2],left_list);
	diff_left+=LabelExist(q3[3],left_list);

	diff_right+=LabelExist(q1[0],right_list);
	diff_right+=LabelExist(q1[1],right_list);
	diff_right+=LabelExist(q2[0],right_list);
	diff_right+=LabelExist(q2[1],right_list);
	diff_right+=LabelExist(q3[0],right_list);
	diff_right+=LabelExist(q3[1],right_list);

	if (diff_left<sig_left)
	{
		sig_left=diff_left;
		sig_right=diff_right;
	}

	//printf("(%d, %d): q(%d,%d,%d,%d)  q(%d,%d,%d,%d)  q(%d,%d,%d,%d)\n",
	//diff_left, diff_right,
	//			             q1[0],q1[1],q1[2],q1[3],
	//			             q2[0],q2[1],q2[2],q2[3],
	//			             q3[0],q3[1],q3[2],q3[3]);

	if ( (sig_left==3)&&(sig_right==4) || ( sig_left==4)&&(sig_right==4))
	  return 0;
	else
	  return 1;

}


//// generate an ultramatrix from the given tree structure.
//// bracketlabel: the label of siblings. start with 1. each time we
//// merge two siblings, we mark that internal node with bracketlabel and
///  the next bracketlabel should be current one plus 1.

int CPreProcessing::GenUltra(UNSIGNCHAR *line, int bracketlabel,int internalindex)
{
	 int loc_close,loc_open;
	 int i=0,j=0,k;
	 int label[2],replace_label,small,big;
	 char replace_str[4],currentword[4];
	UNSIGNCHAR new_line[BUFFERSIZE];
	 int max=0;

	 //// get the first matched brackets,very inside one.
	 loc_close=0;
	 while ( (line[loc_close]!=')') && (line[loc_close]!=0) )
		 loc_close++;
	 if (line[loc_close]==0)
		return max;
	 loc_open=loc_close;

	 while (line[loc_open]!='(')
		loc_open--;

	 /// get the two label numbers inside that matched brackets
	 i=loc_open+1;
	 while (i<loc_close)
	 {
		 if ( (line[i]>='0') && (line[i]<='9') )
		 {
			  k=0;
			  memset(currentword,0,4);
			  while((line[i]>='0') && (line[i]<='9') )
			  {
					 currentword[k++]=line[i];
					 i++;
			  }
			  label[j]=atoi(currentword);
			  if ( (label[j]<100)&&(label[j]>max) )
				 max=label[j];
			  j++;
		  }
		  if (line[i]==')')
			  break;
		   else
			  i++;
	 }
	 if  ( (j>2) || (j<0) )
		 printf("error!,j=%d\n",j);
	 if (j==1)  /// only one label in that braket. should be wrong
	 {
		 printf("only one label=%d\n",label[0]);
		 return max;
	 }

	 //// the correct j should be 2
	 if (label[0]<label[1])
	 {
		 small=label[0];
		 big=label[1];
	 }
	 else
	 {
		 small=label[1];
		 big=label[0];
	 }


	 replace_label=internalindex+1;
	 memset(replace_str,0,4);
	 //itoa(replace_label,replace_str,10);
         sprintf(replace_str,"%d",replace_label); 
          
	 if ( big<100 )  //// the two labels are leaf labels, not a replaced internal node label
	 {
           if ( (ultra[small][big]>0) || (ultra[big][small]>0))
	     printf("exited,ultra[%d][%d]=%d\n",small,big,ultra[small][big]);

           height[replace_label-100] = 1;
           ultra[small][big]=1;
           ultra[big][small]=1;
           childrenlist[replace_label-100][0]=small;
           childrenlist[replace_label-100][1]=big;

	 }
	 else if (small >100) /// the two labels are internal node labels, we need consider all their children leaf nodes
	 {
		 merge(childrenlist[replace_label-100],childrenlist[small-100]);
		 merge(childrenlist[replace_label-100],childrenlist[big-100]);
		 
		 if (height[small-100] > height[big-100])
           height [replace_label-100] = height[small-100]+1;
         else
           height [replace_label-100] = height[big-100]+1;
		/// of these two root nodes
		i=0;k=0;
		//printf("two list labels(%d,%d)\n",small,big);
		while(childrenlist[small-100][i]>0)
		{
			//childrenlist[replace_label-100][k++]=childrenlist[small-100][i];
			j=0;
			while(childrenlist[big-100][j]>0)
			{
				//childrenlist[replace_label-100][k++]=childrenlist[big-100][j];
				//printf("i=%d,j=%d\n",childrenlist[small-100][i],childrenlist[big-100][j]);
				ultra[childrenlist[small-100][i]][childrenlist[big-100][j]]=height [replace_label-100];
				ultra[childrenlist[big-100][j]][childrenlist[small-100][i]]=height [replace_label-100];
				j++;
			}
			i++;
		}

	 }
	 else  /// one leaf node label(small) and another one is a internal node label(big)
	 {

		 j=0;k=0;

		 merge(childrenlist[replace_label-100],small);
		 merge(childrenlist[replace_label-100],childrenlist[big-100]);
        
         height [replace_label-100] = height[big-100]+1;
		 while(childrenlist[big-100][j]>0)
		 {
			 //childrenlist[replace_label-100][k++]=childrenlist[big-100][j];
			 ultra[small][childrenlist[big-100][j]]=height [replace_label-100];
			 ultra[childrenlist[big-100][j]][small]=height [replace_label-100];
			 j++;
		 }
	 }

	 memset (new_line,0,BUFFERSIZE);

	 for (i=0;i<loc_open;i++)
		   new_line[i]=line[i];
	 strcat ((char *)new_line,replace_str);
	 int len=strlen((char *)new_line);
	 for(i=loc_close+1;i<(int)strlen((char *)line);i++)
		 new_line[len++]=line[i];
	 printf("new line=%s\n", new_line);
	 //char ch=getchar();
	 int num=GenUltra(new_line, bracketlabel+1,replace_label);
	 if (num>max)
		max=num;
	 return max;
}


//// put the numbers in source into the dest
void CPreProcessing::merge(int *dest, int *source)
{
  int i,j;
  i=0;
  while(source[i]>0)
  {
	  j=0;
	  while (dest[j]>0)
		if (source[i]==dest[j])
			break;
		else
			j++;
	  if (dest[j]>0)
		 break;
	  else
		 dest[j]=source[i];
	  i++;
  }
}

void CPreProcessing::merge(int *dest, int number)
{
  int i=0;
  while (dest[i]>0)
	if (number==dest[i])
	  break;
	else
	  i++;
  if(dest[i]<=0)
	 dest[i]=number;
}


//// given the input tree structure file(bracket format), generate
//// the ultrametric matrix, then generate the complete quartet
///  set of that tree, and save the quartets to the outfilename.
void CPreProcessing::GenerateSmodelSource(char *filename)
{
  int i,j,k,l;
  int len;
  int count,upper;
  int bufsize=labelNum*(labelNum-1)*(labelNum-2)*(labelNum-3)*3/4;
  char *bigBuf, smallBuf[20];
  int maxHeight=labelNum/2+labelNum%2;
  UNSIGNCHAR q[4];

  FILE *stream;
  if( (stream = fopen( filename, "w+t" )) == NULL )
  {
	  printf( "ERROR!!! create file %s\n", filename );
	  return;
  }
  else
	  printf( "generate source code to %s\n", filename );

  bigBuf= (char *) calloc (bufsize,sizeof(char));


  fputs("%-----------------The matrix entries we need compute-----\n",stream);
  count=0;
  for (i=1;i<labelNum;i++)
  {
	for (j=i+1;j<=labelNum;j++)
	{
		fprintf(stream,"matrix(%d,%d). ",i,j);
		if (count%10==6)
		   fputs("\n",stream);
		count++;
	}
  }
  fputs("\n",stream);
 
  fputs("\n%----Uniqueness constraints: each entry of the matrix has exactly one value-----\n",stream);
  for (i=1;i<=labelNum-1;i++)
	 for (j=i+1;j<=labelNum;j++)
	 {
		 if (domain[j][i]==domain[i][j])
		 {
		    fprintf(stream,"result(%d,%d,%d).\n",i,j,domain[j][i]);
		    continue;
         }
         
         if (domain[i][j]<0) 
            upper=  -domain[i][j]-1;
        else
           upper =domain[i][j];
         
		 memset (bigBuf,0,bufsize);
		 for (k=domain[j][i];k<=upper;k++)
		 {
			 memset(smallBuf,0,20);
			 if (k==upper)
				sprintf(smallBuf,"result(%d,%d,%d)",i,j,k);
			 else
				sprintf(smallBuf,"result(%d,%d,%d), ",i,j,k);
			 strcat(bigBuf,smallBuf);
		  }
		  if (domain[i][j]<0) 
		  {
             memset(smallBuf,0,20);
		     sprintf(smallBuf,",result(%d,%d,%d)",i,j,max_height);
		     strcat(bigBuf,smallBuf);
          }
		  fprintf(stream,"1{%s}1.\n",bigBuf);
	  }


  int lower;
  fputs("\n%-------------equal conditions---------------------------------------\n",stream);
  
  /*for (i=1;i<labelNum-1;i++)
    for (j=i+1;j<labelNum;j++)
      for (k=j+1;k<=labelNum;k++)
      {
            if (domain[i][j]<0)
              domain[i][j]=maxHeight;
            if (domain[i][k]<0)
              domain[i][k]=maxHeight;
            if (domain[j][k]<0)
              domain[j][k]=maxHeight; 
            
            ////[i,j]= [j,k]
            if (domain[j][i]<domain[k][j]) 
               lower = domain[k][j];
            else
               lower=domain[j][i];
               
            if (domain[i][j]<domain[j][k]) 
               upper = domain[i][j];
            else
               upper=domain[j][k];
            for (l=lower;l<=upper;l++)
              fprintf(stream,"equal(%d,%d,%d,%d) :- matrix(%d,%d),result(%d,%d,%d),matrix(%d,%d),result(%d,%d,%d).\n",
                                     i,j,j,k, i,j, i,j,l, j,k, j,k,l);
           
                       ////[i,j]= [i,k]
            if (domain[j][i]<domain[k][i]) 
               lower = domain[k][i];
            else
               lower=domain[j][i];
            if (domain[i][j]<domain[i][k]) 
               upper = domain[i][j];
            else
               upper=domain[i][k];
            for (l=lower;l<=upper;l++)
              fprintf(stream,"equal(%d,%d,%d,%d) :- matrix(%d,%d),result(%d,%d,%d),matrix(%d,%d),result(%d,%d,%d).\n",
                                     i,j,i,k, i,j, i,j,l, i,k, i,k,l);   

                       ////[i,k]= [j,k]
            if (domain[k][i]<domain[k][j]) 
               lower = domain[k][j];
            else
               lower=domain[k][i];
            if (domain[i][k]<domain[j][k]) 
               upper = domain[i][k];
            else
               upper=domain[j][k];
            for (l=lower;l<=upper;l++)
              fprintf(stream,"equal(%d,%d,%d,%d) :- matrix(%d,%d),result(%d,%d,%d),matrix(%d,%d),result(%d,%d,%d).\n",
                                     i,k,j,k, i,k, i,k,l, j,k, j,k,l);                                                                         
               
      }*/
  
  for (i=1;i<=maxHeight;i++)
  {
	  fprintf(stream,"equal(X,Y,Y,Z) :- matrix(X,Y),result(X,Y,%d),matrix(Y,Z),result(Y,Z,%d),X<Y,Y<Z.\n",i,i);
	  fprintf(stream,"equal(X,Y,X,Z) :- matrix(X,Y),result(X,Y,%d),matrix(X,Z),result(X,Z,%d),X<Y,Y<Z.\n",i,i);
	  fprintf(stream,"equal(X,Z,Y,Z) :- matrix(X,Z),result(X,Z,%d),matrix(Y,Z),result(Y,Z,%d),X<Y,Y<Z.\n",i,i);
  }
  
  
  

  fputs("\n%-------------greater conditions---------------------------------------\n",stream);
  for (i=1;i<=maxHeight;i++)
	 for (j=i+1;j<=maxHeight;j++)
	 {
		fprintf(stream,"gter(X,Y,X,Z) :- matrix(X,Y),matrix(X,Z),result(X,Y,%d),result(X,Z,%d),X<Y,Y<Z.\n",j,i);
		fprintf(stream,"gter(X,Y,Y,Z) :- matrix(X,Y),matrix(Y,Z),result(X,Y,%d),result(Y,Z,%d),X<Y,Y<Z.\n",j,i);
		fprintf(stream,"gter(X,Z,X,Y) :- matrix(X,Z),matrix(X,Y),result(X,Z,%d),result(X,Y,%d),X<Y,Y<Z.\n",j,i);
	 }

  /*fputs("\n%------------greater and equal conditions cannot exist for the same two cells------------------\n",stream);
  fprintf(stream," :- matrix(X,A),matrix(X,B),gter(X,A,X,B),equal(X,A,X,B),X<A,X<B.\n");
  fprintf(stream," :- matrix(X,A),matrix(B,X),gter(X,A,B,X),equal(X,A,B,X),X<A,B<X.\n");
  fprintf(stream," :- matrix(A,X),matrix(X,B),gter(A,X,X,B),equal(A,X,X,B),A<X,X<B.\n");
  fprintf(stream," :- matrix(A,X),matrix(B,X),gter(A,X,B,X),equal(A,X,B,X),A<X,B<X.\n");*/


  fputs("\n%-------------ultrametric constraints---------------------------------------\n",stream);
  fputs("ultra(X,Y,Z) :- matrix(X,Y),matrix(Y,Z),matrix(X,Z),equal(X,Y,Y,Z),gter(X,Y,X,Z),X<Y,Y<Z.\n",stream);
  fputs("ultra(X,Y,Z) :- matrix(X,Y),matrix(Y,Z),matrix(X,Z),equal(X,Y,X,Z),gter(X,Y,Y,Z),X<Y,Y<Z.\n",stream);
  fputs("ultra(X,Y,Z) :- matrix(X,Y),matrix(Y,Z),matrix(X,Z),equal(X,Z,Y,Z),gter(X,Z,X,Y),X<Y,Y<Z.\n",stream);



  fputs("\n%-------------quartet consistency constraints---------------------------------------\n",stream);
  memset(bigBuf,0,bufsize);
  count=0;
  
  for (i=1;i<=labelNum-3;i++)
    for (j=i+1;j<=labelNum-2;j++)
      for (k=j+1;k<=labelNum-1;k++)
        for (l=k+1;l<=labelNum;l++)
        {
            count++;
            quartetTable->find(i,j,k,l,q);
            WriteQuartetRule(q[0],q[1],q[2],q[3],stream);
            memset(smallBuf,0,20);
		    if (count%6==0)
	           sprintf(smallBuf,"q(%d,%d,%d,%d)=%d,\n",q[0],q[1],q[2],q[3],quartetTable->GetValue(q[0],q[1],q[2],q[3]));
	       	else
	           sprintf(smallBuf,"q(%d,%d,%d,%d)=%d, " ,q[0],q[1],q[2],q[3],quartetTable->GetValue(q[0],q[1],q[2],q[3]));
	       	strcat(bigBuf,smallBuf);
        }

  fputs("\n%-------------solution--------------------------------------\n",stream);
  if (bigBuf[strlen(bigBuf)]==',')
	bigBuf[strlen(bigBuf)-1]=' ';
  else
	bigBuf[strlen(bigBuf)-2]=' ';

  fprintf(stream,"maximize [%s].\n",bigBuf);
  fprintf(stream,":- [%s]%d.\n",bigBuf,quartetNum-max_k-1);

//  fprintf (stream,"\ncompute 0 {");
  fprintf (stream,"\ncompute  {");
  
  for (i=1;i<=labelNum-2;i++)
	for (j=i+1;j<=labelNum-1;j++)
	  for (k=j+1;k<=labelNum;k++)
	  {
			   fprintf(stream,"ultra(%d,%d,%d)",i,j,k);
			   if ( (k==labelNum)&&(j==labelNum-1) && (i==labelNum-2) )
				  fputs("}.\n",stream);
			   else
				  fputs(", ",stream);
			   if(count%8==0)
				  fputs("\n",stream);
			   count++;

	  }
 fprintf(stream,"\nhide  equal(A,B,C,D), gter(A,B,C,D),matrix(A,B), ultra(A,B,C).\n");

 printf("source codes generated\n");


 fclose(stream);
 free(bigBuf);


}


void CPreProcessing::WriteQuartetRule(int a, int b, int c, int d, FILE *stream)
{
   int big_bc, small_bc;//, small_bd, big_bd;
   if (b>c)
   {
     big_bc=b;
     small_bc=c;
   }
   else
   {
     big_bc=c;
     small_bc=b;
   }

   fprintf(stream,
     "q(%d,%d,%d,%d) :- gter(%d,%d,%d,%d),gter(%d,%d,%d,%d).\n",
     	 a,b,c,d,
     	 a,c,a,b,
     	 a,d,a,b);  //gter(A,C,A,B)

   fprintf(stream,
   	"q(%d,%d,%d,%d) :- gter(%d,%d,%d,%d),gter(%d,%d,%d,%d).\n",
   	 a,b,c,d,
   	 a,c,c,d,   //gter(A,C,C,D)
   	 small_bc,big_bc,c,d);  //gter(B,C,C,D)
}

  
  
  
